@@ -20,14 +20,16 @@ import (
 	"context"
 	"fmt"
 	"reflect"
+	"strconv"
 
 	"go.uber.org/zap"
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
+	"k8s.io/apimachinery/pkg/util/sets"
 	"knative.dev/pkg/apis"
 
-	"github.com/tektoncd/pipeline/pkg/apis/pipeline/v1alpha1"
 	"github.com/tektoncd/pipeline/pkg/apis/pipeline/v1beta1"
+	resourcev1alpha1 "github.com/tektoncd/pipeline/pkg/apis/resource/v1alpha1"
 	"github.com/tektoncd/pipeline/pkg/contexts"
 	"github.com/tektoncd/pipeline/pkg/list"
 	"github.com/tektoncd/pipeline/pkg/names"
@@ -36,39 +38,37 @@ import (
 )
 
 const (
-	// ReasonRunning indicates that the reason for the inprogress status is that the TaskRun
-	// is just starting to be reconciled
-	ReasonRunning = "Running"
-
-	// ReasonFailed indicates that the reason for the failure status is that one of the TaskRuns failed
-	ReasonFailed = "Failed"
-
-	// ReasonCancelled indicates that the reason for the cancelled status is that one of the TaskRuns cancelled
-	ReasonCancelled = "Cancelled"
-
-	// ReasonSucceeded indicates that the reason for the finished status is that all of the TaskRuns
-	// completed successfully
-	ReasonSucceeded = "Succeeded"
-
-	// ReasonCompleted indicates that the reason for the finished status is that all of the TaskRuns
-	// completed successfully but with some conditions checking failed
-	ReasonCompleted = "Completed"
-
-	// ReasonTimedOut indicates that the PipelineRun has taken longer than its configured
-	// timeout
-	ReasonTimedOut = "PipelineRunTimeout"
-
 	// ReasonConditionCheckFailed indicates that the reason for the failure status is that the
 	// condition check associated to the pipeline task evaluated to false
 	ReasonConditionCheckFailed = "ConditionCheckFailed"
 )
 
+// TaskNotFoundError indicates that the resolution failed because a referenced Task couldn't be retrieved
+type TaskNotFoundError struct {
+	Name string
+	Msg  string
+}
+
+func (e *TaskNotFoundError) Error() string {
+	return fmt.Sprintf("Couldn't retrieve Task %q: %s", e.Name, e.Msg)
+}
+
+// ConditionNotFoundError is used to track failures to the
+type ConditionNotFoundError struct {
+	Name string
+	Msg  string
+}
+
+func (e *ConditionNotFoundError) Error() string {
+	return fmt.Sprintf("Couldn't retrieve Condition %q: %s", e.Name, e.Msg)
+}
+
 // ResolvedPipelineRunTask contains a Task and its associated TaskRun, if it
 // exists. TaskRun can be nil to represent there being no TaskRun.
 type ResolvedPipelineRunTask struct {
 	TaskRunName           string
-	TaskRun               *v1alpha1.TaskRun
-	PipelineTask          *v1alpha1.PipelineTask
+	TaskRun               *v1beta1.TaskRun
+	PipelineTask          *v1beta1.PipelineTask
 	ResolvedTaskResources *resources.ResolvedTaskResources
 	// ConditionChecks ~~TaskRuns but for evaling conditions
 	ResolvedConditionChecks TaskConditionCheckState // Could also be a TaskRun or maybe just a Pod?
@@ -78,16 +78,15 @@ type ResolvedPipelineRunTask struct {
 // state of the PipelineRun.
 type PipelineRunState []*ResolvedPipelineRunTask
 
-func (t ResolvedPipelineRunTask) IsDone() (isDone bool) {
+func (t ResolvedPipelineRunTask) IsDone() bool {
 	if t.TaskRun == nil || t.PipelineTask == nil {
-		return
+		return false
 	}
 
 	status := t.TaskRun.Status.GetCondition(apis.ConditionSucceeded)
 	retriesDone := len(t.TaskRun.Status.RetriesStatus)
 	retries := t.PipelineTask.Retries
-	isDone = status.IsTrue() || status.IsFalse() && retriesDone >= retries
-	return
+	return status.IsTrue() || status.IsFalse() && retriesDone >= retries
 }
 
 // IsSuccessful returns true only if the taskrun itself has completed successfully
@@ -125,7 +124,86 @@ func (t ResolvedPipelineRunTask) IsCancelled() bool {
 		return false
 	}
 
-	return c.IsFalse() && c.Reason == v1alpha1.TaskRunSpecStatusCancelled
+	return c.IsFalse() && c.Reason == v1beta1.TaskRunReasonCancelled.String()
+}
+
+// IsStarted returns true only if the PipelineRunTask itself has a TaskRun associated
+func (t ResolvedPipelineRunTask) IsStarted() bool {
+	if t.TaskRun == nil {
+		return false
+	}
+
+	c := t.TaskRun.Status.GetCondition(apis.ConditionSucceeded)
+	if c == nil {
+		return false
+	}
+
+	return true
+}
+
+func (t *ResolvedPipelineRunTask) checkParentsDone(state PipelineRunState, d *dag.Graph) bool {
+	stateMap := state.ToMap()
+	// check if parent tasks are done executing,
+	// if any of the parents is not yet scheduled or still running,
+	// wait for it to complete before evaluating when expressions
+	node := d.Nodes[t.PipelineTask.Name]
+	if isTaskInGraph(t.PipelineTask.Name, d) {
+		for _, p := range node.Prev {
+			if !stateMap[p.Task.HashKey()].IsDone() {
+				return false
+			}
+		}
+	}
+	return true
+}
+
+// Skip returns true if a PipelineTask will not be run because
+// (1) its When Expressions evaluated to false
+// (2) its Condition Checks failed
+// (3) its parent task was skipped
+// (4) Pipeline is in stopping state (one of the PipelineTasks failed)
+// Note that this means Skip returns false if a conditionCheck is in progress
+func (t *ResolvedPipelineRunTask) Skip(state PipelineRunState, d *dag.Graph) bool {
+	// it already has TaskRun associated with it - PipelineTask not skipped
+	if t.IsStarted() {
+		return false
+	}
+
+	// Check if conditionChecks have failed, if so task is skipped
+	if len(t.ResolvedConditionChecks) > 0 {
+		if t.ResolvedConditionChecks.IsDone() && !t.ResolvedConditionChecks.IsSuccess() {
+			return true
+		}
+	}
+
+	// Check if the when expressions are false, based on the input's relationship to the values
+	if t.checkParentsDone(state, d) {
+		if len(t.PipelineTask.WhenExpressions) > 0 {
+			if !t.PipelineTask.WhenExpressions.HaveVariables() {
+				if !t.PipelineTask.WhenExpressions.AllowsExecution() {
+					return true
+				}
+			}
+		}
+	}
+
+	// Skip the PipelineTask if pipeline is in stopping state
+	if isTaskInGraph(t.PipelineTask.Name, d) && state.IsStopping(d) {
+		return true
+	}
+
+	stateMap := state.ToMap()
+	// Recursively look at parent tasks to see if they have been skipped,
+	// if any of the parents have been skipped, skip as well
+	node := d.Nodes[t.PipelineTask.Name]
+	if isTaskInGraph(t.PipelineTask.Name, d) {
+		for _, p := range node.Prev {
+			if stateMap[p.Task.HashKey()].Skip(state, d) {
+				return true
+			}
+		}
+	}
+	return false
 }
 
 // ToMap returns a map that maps pipeline task name to the resolved pipeline run task
@@ -137,23 +215,47 @@ func (state PipelineRunState) ToMap() map[string]*ResolvedPipelineRunTask {
 	return m
 }
 
-func (state PipelineRunState) IsDone() (isDone bool) {
-	isDone = true
+// IsDone returns true when all pipeline tasks have respective taskRun created and
+// that taskRun has either succeeded or failed after all possible retry attempts
+func (state PipelineRunState) IsDone() bool {
 	for _, t := range state {
-		if t.TaskRun == nil || t.PipelineTask == nil {
+		if !t.IsDone() {
 			return false
 		}
-		isDone = isDone && t.IsDone()
-		if !isDone {
-			return
-		}
 	}
-	return
+	return true
 }
 
-// GetNextTasks will return the next ResolvedPipelineRunTasks to execute, which are the ones in the
-// list of candidateTasks which aren't yet indicated in state to be running.
-func (state PipelineRunState) GetNextTasks(candidateTasks map[string]struct{}) []*ResolvedPipelineRunTask {
+// IsBeforeFirstTaskRun returns true if the PipelineRun has not yet started its first TaskRun
+func (state PipelineRunState) IsBeforeFirstTaskRun() bool {
+	for _, t := range state {
+		if t.TaskRun != nil {
+			return false
+		}
+	}
+	return true
+}
+
+// IsStopping returns true if the PipelineRun won't be scheduling any new Task because
+// at least one task already failed or was cancelled in the specified dag
+func (state PipelineRunState) IsStopping(d *dag.Graph) bool {
+	for _, t := range state {
+		if isTaskInGraph(t.PipelineTask.Name, d) {
+			if t.IsCancelled() {
+				return true
+			}
+			if t.IsFailure() {
+				return true
+			}
+		}
+	}
+	return false
+}
+
+// GetNextTasks returns a list of tasks which should be executed next i.e.
+// a list of tasks from candidateTasks which aren't yet indicated in state to be running and
+// a list of cancelled/failed tasks from candidateTasks which haven't exhausted their retries
+func (state PipelineRunState) GetNextTasks(candidateTasks sets.String) []*ResolvedPipelineRunTask {
 	tasks := []*ResolvedPipelineRunTask{}
 	for _, t := range state {
 		if _, ok := candidateTasks[t.PipelineTask.Name]; ok && t.TaskRun == nil {
@@ -162,7 +264,7 @@ func (state PipelineRunState) GetNextTasks(candidateTasks map[string]struct{}) [
 		if _, ok := candidateTasks[t.PipelineTask.Name]; ok && t.TaskRun != nil {
 			status := t.TaskRun.Status.GetCondition(apis.ConditionSucceeded)
 			if status != nil && status.IsFalse() {
-				if !(t.TaskRun.IsCancelled() || status.Reason == v1alpha1.TaskRunSpecStatusCancelled || status.Reason == ReasonConditionCheckFailed) {
+				if !(t.TaskRun.IsCancelled() || status.Reason == v1beta1.TaskRunReasonCancelled.String() || status.Reason == ReasonConditionCheckFailed) {
 					if len(t.TaskRun.Status.RetriesStatus) < t.PipelineTask.Retries {
 						tasks = append(tasks, t)
 					}
@@ -173,31 +275,80 @@ func (state PipelineRunState) GetNextTasks(candidateTasks map[string]struct{}) [
 	return tasks
 }
 
-// SuccessfulPipelineTaskNames returns a list of the names of all of the PipelineTasks in state
-// which have successfully completed.
-func (state PipelineRunState) SuccessfulPipelineTaskNames() []string {
-	done := []string{}
+// SuccessfulOrSkippedDAGTasks returns a list of the names of all of the PipelineTasks in state
+// which have successfully completed or skipped
+func (state PipelineRunState) SuccessfulOrSkippedDAGTasks(d *dag.Graph) []string {
+	tasks := []string{}
 	for _, t := range state {
-		if t.TaskRun != nil {
-			c := t.TaskRun.Status.GetCondition(apis.ConditionSucceeded)
-			if c.IsTrue() {
-				done = append(done, t.PipelineTask.Name)
+		if isTaskInGraph(t.PipelineTask.Name, d) {
+			if t.IsSuccessful() || t.Skip(state, d) {
+				tasks = append(tasks, t.PipelineTask.Name)
 			}
 		}
 	}
-	return done
+	return tasks
+}
+
+// checkTasksDone returns true if all tasks from the specified graph are finished executing
+// a task is considered done if it has failed/succeeded/skipped
+func (state PipelineRunState) checkTasksDone(d *dag.Graph) bool {
+	for _, t := range state {
+		if isTaskInGraph(t.PipelineTask.Name, d) {
+			if t.TaskRun == nil {
+				// this task might have skipped if taskRun is nil
+				// continue and ignore if this task was skipped
+				// skipped task is considered part of done
+				if t.Skip(state, d) {
+					continue
+				}
+				return false
+			}
+			if !t.IsDone() {
+				return false
+			}
+		}
+	}
+	return true
+}
+
+// GetFinalTasks returns a list of final tasks without any taskRun associated with it
+// GetFinalTasks returns final tasks only when all DAG tasks have finished executing successfully or skipped or
+// any one DAG task resulted in failure
+func (state PipelineRunState) GetFinalTasks(d *dag.Graph, dfinally *dag.Graph) []*ResolvedPipelineRunTask {
+	tasks := []*ResolvedPipelineRunTask{}
+	finalCandidates := sets.NewString()
+	// check either pipeline has finished executing all DAG pipelineTasks
+	// or any one of the DAG pipelineTask has failed
+	if state.checkTasksDone(d) {
+		// return list of tasks with all final tasks
+		for _, t := range state {
+			if isTaskInGraph(t.PipelineTask.Name, dfinally) && !t.IsSuccessful() {
+				finalCandidates.Insert(t.PipelineTask.Name)
+			}
+		}
+		tasks = state.GetNextTasks(finalCandidates)
+	}
+	return tasks
+}
+
+// Check if a PipelineTask belongs to the specified Graph
+func isTaskInGraph(pipelineTaskName string, d *dag.Graph) bool {
+	if _, ok := d.Nodes[pipelineTaskName]; ok {
+		return true
+	}
+	return false
 }
 
 // GetTaskRun is a function that will retrieve the TaskRun name.
-type GetTaskRun func(name string) (*v1alpha1.TaskRun, error)
+type GetTaskRun func(name string) (*v1beta1.TaskRun, error)
 
 // GetResourcesFromBindings will retrieve all Resources bound in PipelineRun pr and return a map
 // from the declared name of the PipelineResource (which is how the PipelineResource will
 // be referred to in the PipelineRun) to the PipelineResource, obtained via getResource.
-func GetResourcesFromBindings(pr *v1alpha1.PipelineRun, getResource resources.GetResource) (map[string]*v1alpha1.PipelineResource, error) {
-	rs := map[string]*v1alpha1.PipelineResource{}
+func GetResourcesFromBindings(pr *v1beta1.PipelineRun, getResource resources.GetResource) (map[string]*resourcev1alpha1.PipelineResource, error) {
+	rs := map[string]*resourcev1alpha1.PipelineResource{}
 	for _, resource := range pr.Spec.Resources {
-		r, err := resources.GetResourceFromBinding(&resource, getResource)
+		r, err := resources.GetResourceFromBinding(resource, getResource)
 		if err != nil {
 			return rs, fmt.Errorf("error following resource reference for %s: %w", resource.Name, err)
 		}
@@ -207,7 +358,7 @@ func GetResourcesFromBindings(pr *v1alpha1.PipelineRun, getResource resources.Ge
 }
 
 // ValidateResourceBindings validate that the PipelineResources declared in Pipeline p are bound in PipelineRun.
-func ValidateResourceBindings(p *v1alpha1.PipelineSpec, pr *v1alpha1.PipelineRun) error {
+func ValidateResourceBindings(p *v1beta1.PipelineSpec, pr *v1beta1.PipelineRun) error {
 	required := make([]string, 0, len(p.Resources))
 	optional := make([]string, 0, len(p.Resources))
 	for _, resource := range p.Resources {
@@ -237,8 +388,8 @@ func ValidateResourceBindings(p *v1alpha1.PipelineSpec, pr *v1alpha1.PipelineRun
 }
 
 // ValidateWorkspaceBindings validates that the Workspaces expected by a Pipeline are provided by a PipelineRun.
-func ValidateWorkspaceBindings(p *v1alpha1.PipelineSpec, pr *v1alpha1.PipelineRun) error {
-	pipelineRunWorkspaces := make(map[string]v1alpha1.WorkspaceBinding)
+func ValidateWorkspaceBindings(p *v1beta1.PipelineSpec, pr *v1beta1.PipelineRun) error {
+	pipelineRunWorkspaces := make(map[string]v1beta1.WorkspaceBinding)
 	for _, binding := range pr.Spec.Workspaces {
 		pipelineRunWorkspaces[binding.Name] = binding
 	}
@@ -251,23 +402,34 @@ func ValidateWorkspaceBindings(p *v1alpha1.PipelineSpec, pr *v1alpha1.PipelineRu
 	return nil
 }
 
-// TaskNotFoundError indicates that the resolution failed because a referenced Task couldn't be retrieved
-type TaskNotFoundError struct {
-	Name string
-	Msg  string
+// ValidateTaskRunSpecs that the TaskRunSpecs defined by a PipelineRun are correct.
+func ValidateTaskRunSpecs(p *v1beta1.PipelineSpec, pr *v1beta1.PipelineRun) error {
+	pipelineTasks := make(map[string]string)
+	for _, task := range p.Tasks {
+		pipelineTasks[task.Name] = task.Name
+	}
+
+	for _, taskrunSpec := range pr.Spec.TaskRunSpecs {
+		if _, ok := pipelineTasks[taskrunSpec.PipelineTaskName]; !ok {
+			return fmt.Errorf("PipelineRun's taskrunSpecs defined wrong taskName: %q, does not exist in Pipeline", taskrunSpec.PipelineTaskName)
+		}
+	}
+	return nil
 }
 
-func (e *TaskNotFoundError) Error() string {
-	return fmt.Sprintf("Couldn't retrieve Task %q: %s", e.Name, e.Msg)
-}
+// ValidateServiceaccountMapping validates that the ServiceAccountNames defined by a PipelineRun are not correct.
+func ValidateServiceaccountMapping(p *v1beta1.PipelineSpec, pr *v1beta1.PipelineRun) error {
+	pipelineTasks := make(map[string]string)
+	for _, task := range p.Tasks {
+		pipelineTasks[task.Name] = task.Name
+	}
 
-type ConditionNotFoundError struct {
-	Name string
-	Msg  string
-}
-
-func (e *ConditionNotFoundError) Error() string {
-	return fmt.Sprintf("Couldn't retrieve Condition %q: %s", e.Name, e.Msg)
+	for _, name := range pr.Spec.ServiceAccountNames {
+		if _, ok := pipelineTasks[name.TaskName]; !ok {
+			return fmt.Errorf("PipelineRun's ServiceAccountNames defined wrong taskName: %q, does not exist in Pipeline", name.TaskName)
+		}
+	}
+	return nil
 }
 
 // ResolvePipelineRun retrieves all Tasks instances which are reference by tasks, getting
@@ -276,13 +438,13 @@ func (e *ConditionNotFoundError) Error() string {
 // It will retrieve the Resources needed for the TaskRun using the mapping of providedResources.
 func ResolvePipelineRun(
 	ctx context.Context,
-	pipelineRun v1alpha1.PipelineRun,
+	pipelineRun v1beta1.PipelineRun,
 	getTask resources.GetTask,
 	getTaskRun resources.GetTaskRun,
 	getClusterTask resources.GetClusterTask,
 	getCondition GetCondition,
-	tasks []v1alpha1.PipelineTask,
-	providedResources map[string]*v1alpha1.PipelineResource,
+	tasks []v1beta1.PipelineTask,
+	providedResources map[string]*resourcev1alpha1.PipelineResource,
 ) (PipelineRunState, error) {
 
 	state := []*ResolvedPipelineRunTask{}
@@ -291,20 +453,20 @@ func ResolvePipelineRun(
 
 		rprt := ResolvedPipelineRunTask{
 			PipelineTask: &pt,
-			TaskRunName:  getTaskRunName(pipelineRun.Status.TaskRuns, pt.Name, pipelineRun.Name),
+			TaskRunName:  GetTaskRunName(pipelineRun.Status.TaskRuns, pt.Name, pipelineRun.Name),
 		}
 
 		// Find the Task that this PipelineTask is using
 		var (
-			t        v1alpha1.TaskInterface
+			t        v1beta1.TaskInterface
 			err      error
-			spec     v1alpha1.TaskSpec
+			spec     v1beta1.TaskSpec
 			taskName string
-			kind     v1alpha1.TaskKind
+			kind     v1beta1.TaskKind
 		)
 
 		if pt.TaskRef != nil {
-			if pt.TaskRef.Kind == v1alpha1.ClusterTaskKind {
+			if pt.TaskRef.Kind == v1beta1.ClusterTaskKind {
 				t, err = getClusterTask(pt.TaskRef.Name)
 			} else {
 				t, err = getTask(pt.TaskRef.Name)
@@ -319,12 +481,9 @@ func ResolvePipelineRun(
 			taskName = t.TaskMetadata().Name
 			kind = pt.TaskRef.Kind
 		} else {
-			spec = *pt.TaskSpec
+			spec = *pt.TaskSpec.TaskSpec
 		}
 		spec.SetDefaults(contexts.WithUpgradeViaDefaulting(ctx))
-		if err := spec.ConvertUp(ctx, &v1beta1.TaskSpec{}); err != nil {
-			return nil, err
-		}
 		rtr, err := ResolvePipelineTaskResources(pt, &spec, taskName, kind, providedResources)
 		if err != nil {
 			return nil, fmt.Errorf("couldn't match referenced resources with declared resources: %w", err)
@@ -358,21 +517,21 @@ func ResolvePipelineRun(
 }
 
 // getConditionCheckName should return a unique name for a `ConditionCheck` if one has not already been defined, and the existing one otherwise.
-func getConditionCheckName(taskRunStatus map[string]*v1alpha1.PipelineRunTaskRunStatus, trName, conditionName string) string {
+func getConditionCheckName(taskRunStatus map[string]*v1beta1.PipelineRunTaskRunStatus, trName, conditionRegisterName string) string {
 	trStatus, ok := taskRunStatus[trName]
 	if ok && trStatus.ConditionChecks != nil {
 		for k, v := range trStatus.ConditionChecks {
 			// TODO(1022): Should  we allow multiple conditions of the same type?
-			if conditionName == v.ConditionName {
+			if conditionRegisterName == v.ConditionName {
 				return k
 			}
 		}
 	}
-	return names.SimpleNameGenerator.RestrictLengthWithRandomSuffix(fmt.Sprintf("%s-%s", trName, conditionName))
+	return names.SimpleNameGenerator.RestrictLengthWithRandomSuffix(fmt.Sprintf("%s-%s", trName, conditionRegisterName))
 }
 
-// getTaskRunName should return a unique name for a `TaskRun` if one has not already been defined, and the existing one otherwise.
-func getTaskRunName(taskRunsStatus map[string]*v1alpha1.PipelineRunTaskRunStatus, ptName, prName string) string {
+// GetTaskRunName should return a unique name for a `TaskRun` if one has not already been defined, and the existing one otherwise.
+func GetTaskRunName(taskRunsStatus map[string]*v1beta1.PipelineRunTaskRunStatus, ptName, prName string) string {
 	for k, v := range taskRunsStatus {
 		if v.PipelineTaskName == ptName {
 			return k
@@ -384,119 +543,104 @@ func getTaskRunName(taskRunsStatus map[string]*v1alpha1.PipelineRunTaskRunStatus
 
 // GetPipelineConditionStatus will return the Condition that the PipelineRun prName should be
 // updated with, based on the status of the TaskRuns in state.
-func GetPipelineConditionStatus(pr *v1alpha1.PipelineRun, state PipelineRunState, logger *zap.SugaredLogger, dag *dag.Graph) *apis.Condition {
+func GetPipelineConditionStatus(pr *v1beta1.PipelineRun, state PipelineRunState, logger *zap.SugaredLogger, dag *dag.Graph, dfinally *dag.Graph) *apis.Condition {
 	// We have 4 different states here:
 	// 1. Timed out -> Failed
-	// 2. Any one TaskRun has failed - >Failed. This should change with #1020 and #1023
+	// 2. All tasks are done and at least one has failed or has been cancelled -> Failed
 	// 3. All tasks are done or are skipped (i.e. condition check failed).-> Success
-	// 4. A Task or Condition is running right now  or there are things left to run -> Running
+	// 4. A Task or Condition is running right now or there are things left to run -> Running
 	if pr.IsTimedOut() {
 		return &apis.Condition{
 			Type:    apis.ConditionSucceeded,
 			Status:  corev1.ConditionFalse,
-			Reason:  ReasonTimedOut,
+			Reason:  v1beta1.PipelineRunReasonTimedOut.String(),
 			Message: fmt.Sprintf("PipelineRun %q failed to finish within %q", pr.Name, pr.Spec.Timeout.Duration.String()),
 		}
 	}
 
-	// A single failed task mean we fail the pipeline
-	for _, rprt := range state {
-		if rprt.IsCancelled() {
-			logger.Infof("TaskRun %s is cancelled, so PipelineRun %s is cancelled", rprt.TaskRunName, pr.Name)
-			return &apis.Condition{
-				Type:    apis.ConditionSucceeded,
-				Status:  corev1.ConditionFalse,
-				Reason:  ReasonCancelled,
-				Message: fmt.Sprintf("TaskRun %s has cancelled", rprt.TaskRun.Name),
-			}
-		}
-
-		if rprt.IsFailure() { //IsDone ensures we have crossed the retry limit
-			logger.Infof("TaskRun %s has failed, so PipelineRun %s has failed, retries done: %b", rprt.TaskRunName, pr.Name, len(rprt.TaskRun.Status.RetriesStatus))
-			return &apis.Condition{
-				Type:    apis.ConditionSucceeded,
-				Status:  corev1.ConditionFalse,
-				Reason:  ReasonFailed,
-				Message: fmt.Sprintf("TaskRun %s has failed", rprt.TaskRun.Name),
-			}
-		}
-	}
-
 	allTasks := []string{}
-	successOrSkipTasks := []string{}
+	withStatusTasks := []string{}
 	skipTasks := int(0)
+	failedTasks := int(0)
+	cancelledTasks := int(0)
+	reason := v1beta1.PipelineRunReasonSuccessful.String()
 
 	// Check to see if all tasks are success or skipped
+	//
+	// The completion reason is also calculated here, but it will only be used
+	// if all tasks are completed.
+	//
+	// The pipeline run completion reason is set from the taskrun completion reason
+	// according to the following logic:
+	//
+	// - All successful: ReasonSucceeded
+	// - Some successful, some skipped: ReasonCompleted
+	// - Some cancelled, none failed: ReasonCancelled
+	// - At least one failed: ReasonFailed
 	for _, rprt := range state {
 		allTasks = append(allTasks, rprt.PipelineTask.Name)
-		if rprt.IsSuccessful() {
-			successOrSkipTasks = append(successOrSkipTasks, rprt.PipelineTask.Name)
-		}
-		if isSkipped(rprt, state.ToMap(), dag) {
+		switch {
+		case rprt.IsSuccessful():
+			withStatusTasks = append(withStatusTasks, rprt.PipelineTask.Name)
+		case rprt.Skip(state, dag):
 			skipTasks++
-			successOrSkipTasks = append(successOrSkipTasks, rprt.PipelineTask.Name)
+			withStatusTasks = append(withStatusTasks, rprt.PipelineTask.Name)
+			// At least one is skipped and no failure yet, mark as completed
+			if reason == v1beta1.PipelineRunReasonSuccessful.String() {
+				reason = v1beta1.PipelineRunReasonCompleted.String()
+			}
+		case rprt.IsCancelled():
+			cancelledTasks++
+			withStatusTasks = append(withStatusTasks, rprt.PipelineTask.Name)
+			if reason != v1beta1.PipelineRunReasonFailed.String() {
+				reason = v1beta1.PipelineRunReasonCancelled.String()
+			}
+		case rprt.IsFailure():
+			withStatusTasks = append(withStatusTasks, rprt.PipelineTask.Name)
+			failedTasks++
+			reason = v1beta1.PipelineRunReasonFailed.String()
 		}
 	}
 
-	if reflect.DeepEqual(allTasks, successOrSkipTasks) {
+	if reflect.DeepEqual(allTasks, withStatusTasks) {
+		status := corev1.ConditionTrue
+		if failedTasks > 0 || cancelledTasks > 0 {
+			status = corev1.ConditionFalse
+		}
 		logger.Infof("All TaskRuns have finished for PipelineRun %s so it has finished", pr.Name)
-		reason := ReasonSucceeded
-		if skipTasks != 0 {
-			reason = ReasonCompleted
-		}
-
 		return &apis.Condition{
-			Type:    apis.ConditionSucceeded,
-			Status:  corev1.ConditionTrue,
-			Reason:  reason,
-			Message: fmt.Sprintf("Tasks Completed: %d, Skipped: %d", len(successOrSkipTasks)-skipTasks, skipTasks),
+			Type:   apis.ConditionSucceeded,
+			Status: status,
+			Reason: reason,
+			Message: fmt.Sprintf("Tasks Completed: %d (Failed: %d, Cancelled %d), Skipped: %d",
+				len(allTasks)-skipTasks, failedTasks, cancelledTasks, skipTasks),
 		}
 	}
 
-	// Hasn't timed out; no taskrun failed yet; and not all tasks have finished....
-	// Must keep running then....
+	// Hasn't timed out; not all tasks have finished.... Must keep running then....
+	// transition pipeline into stopping state when one of the tasks(dag/final) cancelled or one of the dag tasks failed
+	// for a pipeline with final tasks, single dag task failure does not transition to interim stopping state
+	// pipeline stays in running state until all final tasks are done before transitioning to failed state
+	if cancelledTasks > 0 || (failedTasks > 0 && state.checkTasksDone(dfinally)) {
+		reason = v1beta1.PipelineRunReasonStopping.String()
+	} else {
+		reason = v1beta1.PipelineRunReasonRunning.String()
+	}
 	return &apis.Condition{
-		Type:    apis.ConditionSucceeded,
-		Status:  corev1.ConditionUnknown,
-		Reason:  ReasonRunning,
-		Message: fmt.Sprintf("Tasks Completed: %d, Incomplete: %d, Skipped: %d", len(successOrSkipTasks)-skipTasks, len(allTasks)-len(successOrSkipTasks), skipTasks),
+		Type:   apis.ConditionSucceeded,
+		Status: corev1.ConditionUnknown,
+		Reason: reason,
+		Message: fmt.Sprintf("Tasks Completed: %d (Failed: %d, Cancelled %d), Incomplete: %d, Skipped: %d",
+			len(withStatusTasks)-skipTasks, failedTasks, cancelledTasks, len(allTasks)-len(withStatusTasks), skipTasks),
 	}
 }
 
-// isSkipped returns true if a Task in a TaskRun will not be run either because
-//  its Condition Checks failed or because one of the parent tasks's conditions failed
-// Note that this means isSkipped returns false if a conditionCheck is in progress
-func isSkipped(rprt *ResolvedPipelineRunTask, stateMap map[string]*ResolvedPipelineRunTask, d *dag.Graph) bool {
-	// Taskrun not skipped if it already exists
-	if rprt.TaskRun != nil {
-		return false
-	}
-
-	// Check if conditionChecks have failed, if so task is skipped
-	if len(rprt.ResolvedConditionChecks) > 0 {
-		// isSkipped is only true iof
-		if rprt.ResolvedConditionChecks.IsDone() && !rprt.ResolvedConditionChecks.IsSuccess() {
-			return true
-		}
-	}
-
-	// Recursively look at parent tasks to see if they have been skipped,
-	// if any of the parents have been skipped, skip as well
-	node := d.Nodes[rprt.PipelineTask.Name]
-	for _, p := range node.Prev {
-		skip := isSkipped(stateMap[p.Task.HashKey()], stateMap, d)
-		if skip {
-			return true
-		}
-	}
-	return false
-}
-
-func resolveConditionChecks(pt *v1alpha1.PipelineTask, taskRunStatus map[string]*v1alpha1.PipelineRunTaskRunStatus, taskRunName string, getTaskRun resources.GetTaskRun, getCondition GetCondition, providedResources map[string]*v1alpha1.PipelineResource) ([]*ResolvedConditionCheck, error) {
+func resolveConditionChecks(pt *v1beta1.PipelineTask, taskRunStatus map[string]*v1beta1.PipelineRunTaskRunStatus, taskRunName string, getTaskRun resources.GetTaskRun, getCondition GetCondition, providedResources map[string]*resourcev1alpha1.PipelineResource) ([]*ResolvedConditionCheck, error) {
 	rccs := []*ResolvedConditionCheck{}
 	for i := range pt.Conditions {
 		ptc := pt.Conditions[i]
 		cName := ptc.ConditionRef
+		crName := fmt.Sprintf("%s-%s", cName, strconv.Itoa(i))
 		c, err := getCondition(cName)
 		if err != nil {
 			return nil, &ConditionNotFoundError{
@@ -504,14 +648,14 @@ func resolveConditionChecks(pt *v1alpha1.PipelineTask, taskRunStatus map[string]
 				Msg:  err.Error(),
 			}
 		}
-		conditionCheckName := getConditionCheckName(taskRunStatus, taskRunName, cName)
+		conditionCheckName := getConditionCheckName(taskRunStatus, taskRunName, crName)
 		cctr, err := getTaskRun(conditionCheckName)
 		if err != nil {
 			if !errors.IsNotFound(err) {
 				return nil, fmt.Errorf("error retrieving ConditionCheck %s for taskRun name %s : %w", conditionCheckName, taskRunName, err)
 			}
 		}
-		conditionResources := map[string]*v1alpha1.PipelineResource{}
+		conditionResources := map[string]*resourcev1alpha1.PipelineResource{}
 		for _, declared := range ptc.Resources {
 			if r, ok := providedResources[declared.Resource]; ok {
 				conditionResources[declared.Name] = r
@@ -525,9 +669,10 @@ func resolveConditionChecks(pt *v1alpha1.PipelineTask, taskRunStatus map[string]
 		}
 
 		rcc := ResolvedConditionCheck{
+			ConditionRegisterName: crName,
 			Condition:             c,
 			ConditionCheckName:    conditionCheckName,
-			ConditionCheck:        v1alpha1.NewConditionCheck(cctr),
+			ConditionCheck:        v1beta1.NewConditionCheck(cctr),
 			PipelineTaskCondition: &ptc,
 			ResolvedResources:     conditionResources,
 		}
@@ -539,13 +684,13 @@ func resolveConditionChecks(pt *v1alpha1.PipelineTask, taskRunStatus map[string]
 
 // ResolvePipelineTaskResources matches PipelineResources referenced by pt inputs and outputs with the
 // providedResources and returns an instance of ResolvedTaskResources.
-func ResolvePipelineTaskResources(pt v1alpha1.PipelineTask, ts *v1alpha1.TaskSpec, taskName string, kind v1alpha1.TaskKind, providedResources map[string]*v1alpha1.PipelineResource) (*resources.ResolvedTaskResources, error) {
+func ResolvePipelineTaskResources(pt v1beta1.PipelineTask, ts *v1beta1.TaskSpec, taskName string, kind v1beta1.TaskKind, providedResources map[string]*resourcev1alpha1.PipelineResource) (*resources.ResolvedTaskResources, error) {
 	rtr := resources.ResolvedTaskResources{
 		TaskName: taskName,
 		TaskSpec: ts,
 		Kind:     kind,
-		Inputs:   map[string]*v1alpha1.PipelineResource{},
-		Outputs:  map[string]*v1alpha1.PipelineResource{},
+		Inputs:   map[string]*resourcev1alpha1.PipelineResource{},
+		Outputs:  map[string]*resourcev1alpha1.PipelineResource{},
 	}
 	if pt.Resources != nil {
 		for _, taskInput := range pt.Resources.Inputs {
